/*******************************************************************************
**	CS5534Lib.c				CS5534 A/D driver file  
**
**  Revised    10/30/07   MJB
**
**  This version uses the QSPI Chip Selects as implemented on the IODP board
**   ADC 1 (channels 1..4) uses PCS1
**   ADC 2 (channels 5..8) uses PCS0
**
**  Copyright 2007  Oceanographic Embedded Systems
**  Release or transfer of this source code is prohibited
**  without the written permisson of OES.
*******************************************************************************/

#include	<cfxbios.h>		// Persistor CF1 System and I/O Definitions
#include	<cfxpico.h>		// Persistor CF1 PicoDOS Definitions  

#include	<assert.h>
#include	<ctype.h>
#include	<errno.h>
#include	<float.h>
#include	<limits.h>
#include	<locale.h>
#include	<math.h> 
#include	<setjmp.h>
#include	<signal.h>
#include	<stdarg.h>
#include	<stddef.h>
#include	<stdio.h> 
#include	<stdlib.h>
#include	"CS5534Lib.h"

#define oesdebug false

//****************************************************
//   This version goes directly  to the QSPI to
//   eliminated Queued PicoBus overhead.
//   It also uses a global pointer to speed the 
//   storage of the data values into the array.
//   The CS5534 also switches the MOSI line back to 
//   an input bit so it can be monitored as as BUSY signal
//****************************************************
#define IMBASE          0xFFFFF000
#define     QMCR            ((uspv) (IMBASE + 0xC00))
#define     QILR            ((ucpv) (IMBASE + 0xC04))
#define     QVLR            ((ucpv) (IMBASE + 0xC05))
#define     QPDR            ((ucpv) (IMBASE + 0xC15))
#define     QPAR            ((ucpv) (IMBASE + 0xC16))
#define     QDDR            ((ucpv) (IMBASE + 0xC17))
#define     SPCR0           ((uspv) (IMBASE + 0xC18))
#define     SPCR1           ((uspv) (IMBASE + 0xC1A))
#define     SPCR2           ((uspv) (IMBASE + 0xC1C))
#define     SPCR3           ((ucpv) (IMBASE + 0xC1E))
#define     SPSR            ((ucpv) (IMBASE + 0xC1F))
#define     SPICMD          ((ucpv) (IMBASE + 0xD40))
#define     SPICMDARRAY     ((uchar *) (IMBASE + 0xD40))


typedef volatile uchar      *ucpv;
typedef unsigned int        unsint;
typedef volatile ushort     *uspv;
typedef volatile ulong      *ulpv;

short spdidx = 0;

// prototypes for internal functions


void Send8(short chip,ushort cmd);
void Send32(short chip, ulong data);
long Read32(short chip);
void WaitData(short chip);





// numbering of physical channels is set up as normal for IODP board
// 
ushort setups[16] = { 0x0000,0x4000,0x8000,0xc000,
					0x0000,0x0000,0x0000,0x0000, 
					0x0000,0x4000,0x8000,0xC000,
					0x0000,0x0000,0x0000,0x0000};

const short spdbits[10] = {4,3,2,1,0,12,11,10,9,8};
const short bit2idx[16] = {4,3,2,1,0,10,10,10,9,8,7,6,5,10,10,10};

const float speeds[11] = {7.5,15,30,60,120,240,480,960,1920,3840, 0.0};

ushort CSel1 =   0x0D;   // only PCS1 low
ushort CSel2 =   0x0E;   // only PCS0 low
uchar  ADSlot1   = 0;
uchar  ADSlot2   = 1;   // default QSPI slots

uspv cmdbyteptr;	// location to store command byte to transmit
ulpv xmtdatptr;		// address of long word to transmit A/D
ulpv rcvdatptr;		// address of long word returned from A/D

ucpv  cmdptr1;
ucpv  cmdptr2; // QSPI command RAM addresses


static unsigned char cfqpdr,cfqpar,cfqddr;
static unsigned int  cfspcr0,cfspcr1,cfspcr2,cfspcr3;


// This routine is called automatically in CS5534Init
void CS5534SaveCFQSM(void){
	
	cfqpdr = *QPDR;
	cfqpar = *QPAR;
	cfqddr = *QDDR;
	
	cfspcr0 = *SPCR0;
	cfspcr1 = *SPCR1;
	cfspcr2 = *SPCR2;
	cfspcr3 = *SPCR3;

}

void CS5534RestoreCFQSM(void){
     *SPCR1 = 0x0000;  // disable the qspi while restoring registers 
 	 *QPDR = cfqpdr;
 	 *QPAR = cfqpar;
 	 *QDDR = cfqddr;
	
 	 *SPCR0 = cfspcr0;
	 *SPCR2 = cfspcr2;
 	 *SPCR3 = cfspcr3;	
 	 *SPCR1 = cfspcr1 ; // write last, because it could enable

}

// Set up the QSM using direct hardware access
// also set up our copies of the registers
void CS5534SetupQSM(ushort slot1){ 	// set up as master for A/D I/O 
ushort slot2;
	*SPCR1 = 0x000;		// disable the SPI
	*QPDR = 0xFB;		// set , SCK low when not used 
	*QPAR = 0xFF;  		// set all as QSM pins 
	*QDDR = 0xFE; 		// set pin 0, as input 
    slot2 = slot1+1;

	*SPCR0 = 0x8006; 	// sets master, 16 bits  baud divisor 16, CPHA 1
						
	*SPCR2 = 0;			// Setting NEWQP in SPCR2 will cause a transaction
	*SPCR3 = 0; 		// halt not enabled, no loop back
	SPICMDARRAY[0] = 0; // default transfer  no chip selects set
	*SPCR1 = 0x8000;    // Enable now
	Delay2us();
	while(*SPSR < 128){ }; // wait until qsm done flag set

// we always use 2 sets of QSPI registers starting at slot1	value
// which is normally 9 for demo programs
	cmdbyteptr  =	((ushort*) (IMBASE + 0xD20 + (slot1 *2)));
	xmtdatptr 	= 	((ulong *) (IMBASE + 0xD20 + (slot1 *2)));
	rcvdatptr 	=	((ulong*)  (IMBASE + 0xD00 + (slot1 *2)));
	ADSlot1 = slot1;
	ADSlot2 = slot2;
	cmdptr1 = &SPICMDARRAY[slot1];	
	cmdptr2 = &SPICMDARRAY[slot2];	

	
}


// set up command pointer to send just one byte 
void Send8(short chip,ushort cmd){
	if(chip == 0){
		*cmdptr1 = 0x080+CSel1;
	}  else {
		*cmdptr1 = 0x080+CSel2;
	}
	*cmdbyteptr = cmd;	
	while(*SPSR < 128){ }; // wait until qsm done flag set
	
	*SPCR2 =(ADSlot1<<8) + ADSlot1;
	*SPSR = 0;
	*SPCR0 = 0x8006; 	// sets master, 16 bits  baud divisor 16, CPHA 1
	*SPCR1 |= 0x8000;   // start qsm to get last result
	Delay2us();
	while(*SPSR < 128){ }; // wait until qsm done flag set

}

void Send32(short chip, ulong data){
	while(*SPSR < 128){ }; // wait until qsm done flag set
	if(chip == 0){
		*cmdptr1 = 0x0C0+CSel1;
		*cmdptr2 = 0x040+CSel1;
	}  else {
		*cmdptr1 = 0x0C0+CSel2;
		*cmdptr2 = 0x040+CSel2;
	}
	*xmtdatptr = data;
	*SPCR2 = (ADSlot2<<8) + ADSlot1;   	// set endqp and newqp

	*SPCR0 = 0x8006; 	// sets master, 16 bits  baud divisor 16, CPHA 1?????

	*SPSR = 0;
	*SPCR1 |= 0x8000;   // start qsm to get last result
	Delay2us();
	while(*SPSR < 128){ }; // wait until qsm done flag set

}
void ShowQSPI(void){


	cprintf(" SPCRO: %04X  ",*SPCR0);
	cprintf(" SPCR1: %04X  ",*SPCR1);
	cprintf(" SPCR2: %04X  ",*SPCR2);
	cprintf(" SPCR3: %04X\n",*SPCR3);


}

long Read32(short chip){
	register long rdata;
	while(*SPSR < 128){ }; // wait until qsm done flag set
	if(chip == 0){
		*cmdptr1 = 0xC0+CSel1;
		*cmdptr2 = 0x40+CSel1;
	}  else {
		*cmdptr1 = 0xC0+CSel2;
		*cmdptr2 = 0x40+CSel2;
	}
	*SPCR2 = (ADSlot2<<8) + ADSlot1;   	// set endqp and newqp
	*SPSR = 0;
	*SPCR1 |= 0x8000;   // start qsm to get last result
	Delay2us();
	while(*SPSR < 128){ }; // wait until qsm done flag set


	rdata = *rcvdatptr;   	 

	return rdata;
}
 

void CS5534Init(ushort slot1){
	ulong rval;
	ushort slot2;
	
	slot2 = slot1+1;
	CS5534SaveCFQSM();   // save the state of QSM registers set by CFX
	CS5534SetupQSM(slot1);
	CS5534SerialReset();
	
	
	Send8(0,WRITECONFIG);
	Send32(0,0x20000000);  // only the reset bit is high
	RTCDelayTicks(400);
	
	Send8(0,WRITECONFIG); 
	Send32(0,0x02000000);  // VRS bit is high

	RTCDelayTicks(200);
	Send8(0,READCONFIG);

	rval = Read32(0);
#if oesdebug
	cprintf("\nConfig Register A/D 1 = %08lX \n",rval);
#endif	
	Send8(1,WRITECONFIG);
	Send32(1,0x20000000);  // only the reset bit is high
	RTCDelayTicks(400);
	
	Send8(1,WRITECONFIG); 
	Send32(1,0x02000000);  // VRS bit is high

	RTCDelayTicks(200);
	Send8(1,READCONFIG);

	rval = Read32(1);
#if oesdebug
	cprintf("Config Register A/D 2 = %08lX \n",rval);
#endif
		
}



// wait for Data Ready on channel
// This routine changes the QSPI chip selects back
// to I/O bits so that they can be held stable
void WaitData(short chip){
	ushort pval;
	
 	RTCDelayTicks(8000L); // wait 200msec
 	return;
	if(chip == 0){
		pval = (CSel1 * 8);//  0x0D * 8 = 0x68
	} else {
		pval  = (CSel2 * 8);  // 0x0e * 8 = 0x70
	}

	*QPDR = pval;  // set data output register

	*QPAR = 0x80;  		// set 0..6 as I/O leave 7 as TXD


	while((*QPDR) & 0x01){};  // wait for (MISO) data pin to go low

	*QPAR = 0xFF;  		// set all as QSM pins 
	//PinClear(1);

}


void CS5534RequestData(short chan){

	uchar cch;
	short chip= 0;

	*SPCR1 = 0x000;		// disable the SPI
	*QPDR = 0xFB;		// set , SCK low when not used 
	*QPAR = 0xFF;  		// set all as QSM pins 
	*QDDR = 0xFE; 		// set pin 0, as input 

	chan--;
	if((chan <0) ||(chan >7)) return;
	cch = CONVERT1 + ((chan & 0x03) <<3);
	if(chan > 3) chip = 1;
	Send8(chip, cch);
	Delay2us();  // give ADC time to act on request

}

bool CS5534DataAvailable(short chan){
	short chip;
	ushort pval;
	bool bval;	
	//uchar regval;
	while(*SPSR < 128){ }; // wait until qsm done flag set	
	chan--;
	chip = 0;
	if((chan <0) || (chan > 7)) return false;
	
	if(chan > 3){
		chip = 1;
	}
	

	
	if(chip == 0){
		pval = (CSel1 * 8) ;
	} else {
		pval  = (CSel2 * 8);
	}

	*QPDR = pval;
//	regval = *QPAR;   // get current value
	*QPAR = 0x80;  		// set 0..6 as I/O leave 7 as TXD
	
	Delay2us();  // allow some response time

	bval = (((*QPDR) & 0x01) == 0);  // data ready when pin is low

	//*QPAR = regval;  		// restore QSM pins 
	*QPAR = 0xFF;  		// set all as QSM pins 
	return bval;
	
	
 

}

long CS5534Result(short chan){
		short chip= 0;
	long rval;
	bool unipolar;

	*SPCR1 = 0x000;		// disable the SPI
	*QPDR = 0xFB;		// set , SCK low when not used 
	*QPAR = 0xFF;  		// set all as QSM pins 
	*QDDR = 0xFE; 		// set pin 0, as input 
	chan--;
	if((chan <0) ||(chan >7)) return(0xFFFF0000);
	if(chan > 3) chip = 1;
	unipolar = (((setups[chan + chip * 4]) & 0x0040) == 0x0040);
	Send8(chip, READDATA);
	
	rval = Read32(chip);
	rval = rval >> 8;
	if(unipolar) rval &= 0x00FFFFFF;

	return rval;
}



// Serial reset of both chips
void CS5534SerialReset(void){
	short i;

	for(i=0; i<15; i++){
		Send8(0, 0xFF);  // 15 0xFFs
	}
	Send8(0, 0xFE);  // 1 0xFE
	
	for(i=0; i<15; i++){
		Send8(1, 0xFF);  // 15 0xFFs
	}
	Send8(1, 0xFE);  // 1 0xFE
}


// This routine writes back to the A/Ds the last valid
// data that was written to the channel setup registers
// It is used after low-power sleep may have messed up
// the A/D registers
void CS5534RestoreSetup(void){

	ulong *setptr;
	// now write the setup register array
	Send8(0,0x45);
	setptr = &setups[0];
	Send32(0, *setptr++);  // setup register 0,1
	Send32(0, *setptr++);  // setup register 2,3
	Send32(0, *setptr++);  // setup register 
	Send32(0, *setptr++);  // setup register 
	
	// now do the same with A/D #2 using setups 8..11
	Send8(1,0x45);
	setptr = &setups[8];
	Send32(1, *setptr++);  // setup register 0,1
	Send32(1, *setptr++);  // setup register 2,3
	Send32(1, *setptr++);  // setup register 
	Send32(1, *setptr++);  // setup register 	
}

void SetAllSpeeds( short sdx){

	ulong *setptr;
	ushort spbits;
	
	spdidx = sdx;
	spbits = spdbits[sdx]<<7;

#if oesdebug
	printf("Setting all conversion speed to %5.2f \n", speeds[sdx]);
	
#endif
	setups[0] = (setups[0] & 0xF87F) | spbits;;  //    channel 0
	setups[1] = (setups[1] & 0xF87F) | spbits;;  //    channel 1
	setups[2] = (setups[2] & 0xF87F) | spbits;;  //    channel 2
	setups[3] = (setups[3] & 0xF87F) | spbits;;  //    channel 3
	
	setups[8] = (setups[8] & 0xF87F) | spbits;;  //    channel 0
	setups[9] = (setups[9] & 0xF87F) | spbits;;  //    channel 1
	setups[10] = (setups[10] & 0xF87F) | spbits;;  //    channel 2
	setups[11] = (setups[11] & 0xF87F) | spbits;;  //    channel 3
	
	// now write the setup register array
	Send8(0,0x45);
	setptr = &setups[0];
	Send32(0, *setptr++);  // setup register 0,1
	Send32(0, *setptr++);  // setup register 2,3
	Send32(0, *setptr++);  // setup register 
	Send32(0, *setptr++);  // setup register 
	
	// now do the same with A/D #2 using setups 8..11
	Send8(1,0x45);
	setptr = &setups[8];
	Send32(1, *setptr++);  // setup register 0,1
	Send32(1, *setptr++);  // setup register 2,3
	Send32(1, *setptr++);  // setup register 
	Send32(1, *setptr++);  // setup register 
	
}

void SetSpeed(short chan, short sdx){

	ulong *setptr;


	ushort spbits, chip;
	spbits = spdbits[sdx]<<7;
	chan--;
	chip = 0;
	if((chan <0) || (chan > 7)) return;
	
	if(chan > 3){
		chan+=4;
		chip = 1;
	}
	setups[chan] = (setups[chan] & 0xF87F) | spbits;
	setptr = &setups[chip * 8];

	// now write the proper  setup register array
	Send8(chip,0x45);
	Send32(chip, *setptr++);  // setup register 0,1
	Send32(chip, *setptr++);  // setup register 2,3  first A/D
	Send32(chip, *setptr++);  // setup register  unused
	Send32(chip, *setptr++);  // setup register  unused

}

void SetBipolar(short chan, bool bipolar){

	ulong *setptr;
	ushort bpmask = 0;
	ushort chip;
	chan--;
	chip = 0;
	if((chan <0) || (chan > 7)) return;
	if(!bipolar) bpmask = 0x0040;
	
	if(chan > 3){
		chan+=4;
		chip = 1;
	}
	setups[chan] = (setups[chan] & 0xFFBF) | bpmask;
	setptr = &setups[chip * 8];

	// now write the proper  setup register array
	Send8(chip,0x45);
	Send32(chip, *setptr++);  // setup register 0,1
	Send32(chip, *setptr++);  // setup register 2,3  first A/D
	Send32(chip, *setptr++);  // setup register   unused
	Send32(chip, *setptr++);  // setup register   unused

}

void ShowSetup(void){
	short i, gval,rtidx;
	ushort rvals[8];
	ulong *lptr;
	bool unipolar;
	cprintf("\n");
	Send8(0,0x0D);  // read setup register 0
	lptr = &rvals[0];
	*lptr++ = Read32(0);
	Send8(0,0x1D);  // read setup register 1
	*lptr++ = Read32(0);

// now do the second A/D
	Send8(1,0x0D);  // read setup register 0
	*lptr++ = Read32(1);
	
	Send8(1,0x1D);  // read setup register 1
	*lptr = Read32(1);
	cprintf("\nSetup Register Values\n");
	for(i= 0; i<8; i++){
		cprintf("Ch.#%d  [0x%04X]",i+1, rvals[i]);
		cprintf(" chip:%d ", i/4);
		cprintf(" pchan:%d ", rvals[i]>>14);
		gval = 1 << ((rvals[i]>>11) &0x0007);
		cprintf(" gain:%2d  ", gval);
		rtidx = bit2idx[(rvals[i] & 0x0780)>>7];
		printf(" rate:%7.1f  ", speeds[rtidx]);
		unipolar = (((rvals[i]) & 0x0040) == 0x0040);
		if(unipolar) printf(" Unipolar\n"); else printf(" Bipolar\n");
	}
}



long CS5534GetValue(short chan){

	short chip= 0;
	long rval;



	if((chan <1) ||(chan >8)) return(0xFFFF0000);
	//  CS5534RequestData and CS5534Result will decrement channel numbers
	CS5534RequestData(chan);

	RTCDelayMicroSeconds(1000000l);  // wait one second
	
	rval = CS5534Result(chan);
#if oesdebug
	cprintf(" %08lX ", rval);
#endif
	return rval;
}

void CS5534SetPower(short level, short chip){

	//Remove PowerDown (Permanently Turn off) and Add PowerSave (Very Low Power) - K. Darnell 04/10

	//if(level == POWERDOWN){
		//if(chip == 0){
			//Send8(0,WRITECONFIG); 
			//Send32(0,0xC2000000);  // chip 1  Power Down, Power Save, VRS bits high
		//}
		//if(chip == 1){		
			//Send8(1,WRITECONFIG); 
			//Send32(1,0xC2000000);  // chip 2  Power Down, Power Save, VRS bits high
		//}
	//}
	if(level == POWERNORMAL){
		if(chip == 0){
			Send8(0,WRITECONFIG); 
			Send32(0,0x02000000);  // chip 1 Power Down, Power Save, VRS bits low
		}
		if(chip == 1){
			Send8(1,WRITECONFIG); 
			Send32(1,0x02000000);  // chip 2 Power Down, Power Save, VRS bits low
		}
	}	
	if(level == POWERSAVE){
		if(chip == 0){
			Send8(0,WRITECONFIG); 
			Send32(0,0x42000000);  // chip 1 Power Save
		}
		if(chip == 1){
			Send8(1,WRITECONFIG); 
			Send32(1,0x42000000);  // chip 2 Power Save
		}
	}
	
}


void SetGain(short chan, ushort gainval){
    ushort gv,idx = 0; 
	ulong *setptr;
	ushort chip; 	
    gv = gainval;

 
	
  	while(gv > 1){
  		idx++;
  		gv /= 2;	
  	}
	idx &= 0x07;

	chan--;
	chip = 0;
	if((chan <0) || (chan > 7)) return;
	
	if(chan > 3){
		chan+=4;
		chip = 1;
	}
	setups[chan] = (setups[chan] & 0xC7FF) | (idx<<11);
	setptr = &setups[chip * 8];

	// now write the proper  setup register array
	Send8(chip,0x45);
	Send32(chip, *setptr++);  // setup register 0,1
	Send32(chip, *setptr++);  // setup register 2,3  first A/D
	Send32(chip, *setptr++);  // setup register   unused
	Send32(chip, *setptr++);  // setup register   unused

}


void ShortInputs(void){
	ulong rval;
	
	Send8(0,WRITECONFIG);
	Send32(0, 0x0A000000);  // Short input and VRS bits  high

	RTCDelayTicks(200);
	Send8(0,READCONFIG);
	rval = Read32(0);
#if oesdebug
	cprintf("\nConfig Register Chip 0 = %08lX ",rval);
#endif	
	
	
	Send8(1,WRITECONFIG);
	Send32(1, 0x0A000000);  // Short input and VRS bits  high

	RTCDelayTicks(200);
	Send8(1,READCONFIG);

	rval = Read32(1);
#if oesdebug
	cprintf("\nConfig Register Chip 1 = %08lX \n",rval);
#endif

}


void CS5534SetGainCal(short chan, ulong gaincal){

	short setidx,hdwchan;
	short  chip= 0;

	chan--;
	
	if((chan > 7) || (chan <0)) return ;
	
	setidx = chan;
	if(chan > 3){
		chip = 1;
		setidx += 4;
	}
	hdwchan = setups[setidx] >> 14;

		
	// when writing the offset calibration register, we send
	// the actual hardware channel---not the register pointer
	
	// now write the gain calibration value
	Send8(chip, (hdwchan<<4) | 0x02);  // write gain cal register 
	Send32(chip, gaincal); 
}

void CS5534SetOffCal(short chan, ulong offcal){

	short setidx,hdwchan;
	short  chip= 0;

	chan--;
	
	if((chan > 7) || (chan <0)) return ;
	
	setidx = chan;
	if(chan > 3){
		chip = 1;
		setidx += 4;
	}
	hdwchan = setups[setidx] >> 14;

		
	// when writing the offset calibration register, we send
	// the actual hardware channel---not the register pointer
	
	// now write the offset value
	Send8(chip, (hdwchan<<4) | 0x01);  // write offset register 
	Send32(chip, offcal);  
}


ulong CS5534GetOffCal(short chan){
	ulong rval;
	short setidx,hdwchan;
	short  chip= 0;

	chan--;
	
	if((chan > 7) || (chan <0)) return 0;
	
	setidx = chan;
	if(chan > 3){
		chip = 1;
		setidx += 4;
	}
	hdwchan = setups[setidx] >> 14;

		
	// when reading the offset calibration register, we send
	// the actual hardware channel---not the register pointer
	

	Send8(chip, (hdwchan<<4) | 0x09);  // read offset register 
	rval = Read32(chip);

	return rval;
}


ulong CS5534GetGainCal(short chan){
	ulong rval;
	short setidx,hdwchan;
	short  chip= 0;

	chan--;
	
	if((chan > 7) || (chan <0)) return  0L;
	
	setidx = chan;
	if(chan > 3){
		chip = 1;
		setidx += 4;
	}
	chan &= 0x03;   // mask for values 0..3
	
	hdwchan = setups[setidx] >> 14;
		
	// when reading the gain calibration register, we send
	// the actual hardware channel---not the register pointer

	Send8(chip, (hdwchan<<4) | 0x0A);  // read Gain register 
	rval = Read32(chip);

    return rval;
}



